///////////////////////////////////////////////////////////////////////////////
//                                                         
// CegoBTreeManager.cc
// -------------------
// Cego index manager class implementation
//                                                         
// Design and Implementation by Bjoern Lemke               
//         
// (C)opyright 2000-2014 Bjoern Lemke
//
// IMPLEMENTATION MODULE
//
// Class: CegoBTreeManager
// 
// Description: This class implements the btree algorithm to be used as an alternative index object.
//
//
///////////////////////////////////////////////////////////////////////////////

// base includes
#include <lfcbase/Exception.h>

// cego includes
#include "CegoBTreeManager.h"
#include "CegoLockHandler.h"
#include "CegoTupleState.h"

#include <string.h>
#include <stdlib.h>

CegoBTreeManager::CegoBTreeManager(CegoObjectManager *pObjMng, CegoBTreeObject *pBTO)
{
    _pDBMng = pObjMng->getDBMng();
    _modId = _pDBMng->getModId("CegoBTreeManager");
    _pObjMng = pObjMng;
    _pBTO = pBTO;

    _tabSetId = _pBTO->getTabSetId();
    _btreeName = _pBTO->getName();    
    _btreeType = _pBTO->getType();
    _btreeSchema = _pBTO->getSchema(); 

    CegoField *pF = _btreeSchema.First();
    _keyLen = 0;
    while ( pF )
    {
	_keyLen += pF->getLength();
	pF = _btreeSchema.Next();
    }
    
    int fileId = _pBTO->getDataFileId();
    int pageId = _pBTO->getDataPageId();

    if ( fileId == 0 && pageId == 0 )
    {
	throw Exception(EXLOC, Chain("BTree object is not valid"));
    }

    _rootFixed = false;
    
}

CegoBTreeManager::~CegoBTreeManager()
{
    if ( _rootFixed )
	unfixRoot();
}

void CegoBTreeManager::fixRoot()
{
    int fileId = _pBTO->getDataFileId();
    int pageId = _pBTO->getDataPageId();
   
    // cout << "Fixing root page " << fileId << " " << pageId << endl;
    _pDBMng->bufferFix(_rootBP, _tabSetId, fileId, pageId, CegoBufferPool::NOSYNC, _pObjMng->getLockHandler());
    _rootFixed = true;
    
}

void CegoBTreeManager::unfixRoot()
{
    // cout << "Unfixing root .." << _rootBP.getFileId() << " "  << _rootBP.getPageId() << endl;
    _pDBMng->bufferUnfix(_rootBP, true, _pObjMng->getLockHandler());    
    _rootFixed = false;
    // cout << "Unfixed .." << endl;    
}


void CegoBTreeManager::insertBTree(const CegoDataPointer& dp,
				   const CegoBTreeValue &iv,
				   int tid)
{

    
    // we have to keep the index object in buffer pool 
    CegoBufferPage bp;
    _pObjMng->getObjectWithFix(_tabSetId, _btreeName, _btreeType, *_pBTO, bp);
    
    CegoDataPointer sysEntry(bp.getFileId(), bp.getPageId(), bp.getEntryPos());

    try 
    {
	insertBTree(sysEntry, dp, iv, tid);
    }
    catch ( Exception e )
    {
	_pDBMng->bufferUnfix(bp, true, _pObjMng->getLockHandler());
	throw e;
    }
    _pDBMng->bufferUnfix(bp, true, _pObjMng->getLockHandler());
    
}

void CegoBTreeManager::insertBTree(const CegoDataPointer& sysEntry, 
				   const CegoDataPointer& dp,
				   const CegoBTreeValue &iv,
				   int tid)
{

    bool doAppend = true;

    StackT<CegoBufferPage> parentPageStack;
    StackT<CegoBufferPage> fixedPageStack;

    StackT<CegoBTreeNode> modNodeStack;

    try
    {

	int fileId=0;
	int pageId=0;

	if ( _rootFixed == false )
	{
	    fileId = _pBTO->getDataFileId();
	    pageId = _pBTO->getDataPageId();	    
	}
	    
	bool isInserted = false;

	bool isFirst = true;

	while ( ! isInserted )
	{
	    
	    CegoBufferPage bp;
	    
	    if ( isFirst && _rootFixed)
	    {
		bp = _rootBP;
	    }
	    else
	    {
		_pDBMng->bufferFix(bp, _tabSetId, fileId, pageId, CegoBufferPool::NOSYNC, _pObjMng->getLockHandler());
	    }
	    
	    if ( bp.getType() == CegoBufferPage::BTREE_NODE )
	    {
		
		CegoBTreeNode traceNode;
		traceNode.setType(CegoBTreeNode::NODE);
		traceNode.setPtr(bp.getChunkEntry(), bp.getChunkLen());
		traceNode.setSchema(&_btreeSchema, _keyLen);
		traceNode.setFileId(bp.getFileId());
		traceNode.setPageId(bp.getPageId());

		// traceNode.printNode();

		// cout << "Tracing node " << bp.getFileId() << " " << bp.getPageId() << endl; 
		traceNode.getChildPage(iv, fileId, pageId);
		
		// cout << "Tracing to child page .." << fileId << " " << pageId << endl; 
		parentPageStack.Push(bp);
	    }
	    else if ( bp.getType() == CegoBufferPage::BTREE_LEAF )
	    {
		
		CegoBTreeNode leafLeft;
		leafLeft.setType(CegoBTreeNode::LEAF);
		leafLeft.setPtr(bp.getChunkEntry(), bp.getChunkLen());
		leafLeft.setFileId(bp.getFileId());
		leafLeft.setPageId(bp.getPageId());
		leafLeft.setSchema(&_btreeSchema, _keyLen);
		leafLeft.setNextFileId(bp.getNextFileId());
		leafLeft.setNextPageId(bp.getNextPageId());
		
		bool duplicateAdded;
		if ( leafLeft.addValue(iv, dp, _btreeName, _btreeType == CegoObject::PBTREE || _btreeType == CegoObject::UBTREE, _pObjMng, _tabSetId, tid, duplicateAdded) == false )
		{
		    
		    // cout << "Creating new leaf page .." << endl;
		    CegoBTreeNode leafRight;
		    
		    // create new 
		    CegoBufferPage newPage;
		    _pObjMng->getNewFilePage(newPage, _tabSetId, CegoObject::BTREE, true, doAppend);
		    
		    newPage.initPage(CegoBufferPage::BTREE_LEAF);
		    fixedPageStack.Push(newPage);
		    
		    leafRight.setType(CegoBTreeNode::LEAF);
		    leafRight.setPtr(newPage.getChunkEntry(), newPage.getChunkLen());
		    leafRight.initNode();
		    leafRight.setFileId(newPage.getFileId());
		    leafRight.setPageId(newPage.getPageId());
		    leafRight.setSchema(&_btreeSchema, _keyLen);
		    
		    // cout << "Splitting leaf page .." << endl;
		    leafLeft.split(leafRight);
		    
		    // cout << "Setting next page to " << newPage.getFileId() << " " << newPage.getPageId() << endl;
		    	
		    if ( bp != _rootBP )
			fixedPageStack.Push(bp);
		    
		    CegoBTreeValue ri = leafRight.getMin();
		    
		    bool dupAdded;
		    if ( iv.isHigher(ri, &_btreeSchema) )
		    {
			leafRight.addValue(iv, dp, _btreeName, _btreeType == CegoObject::PBTREE || _btreeType == CegoObject::UBTREE, _pObjMng, _tabSetId, tid, dupAdded);
		    }
		    else
		    {
			leafLeft.addValue(iv, dp, _btreeName, _btreeType == CegoObject::PBTREE || _btreeType == CegoObject::UBTREE, _pObjMng, _tabSetId, tid, dupAdded);
		    }
		    
		    CegoBTreeValue propValue = leafLeft.getMax();
		    CegoBTreeNode propRight = leafRight;
		    CegoBTreeNode propLeft = leafLeft;
			    		    
		    bool isPropagated = false;
		    while ( ! isPropagated )
		    {
			// propagate newLeaf to parent node
			CegoBufferPage propPage;
			
			if ( parentPageStack.Pop(propPage) )
			{

			    // cout << "Adding propPage to stack" << propPage.getFileId() << " "  << propPage.getPageId() << endl;
			    fixedPageStack.Push(propPage);
			    
			    CegoBTreeNode nodeLeft;
			    nodeLeft.setType(CegoBTreeNode::NODE);
			    nodeLeft.setPtr(propPage.getChunkEntry(), propPage.getChunkLen());
			    nodeLeft.setFileId(propPage.getFileId());
			    nodeLeft.setPageId(propPage.getPageId());
			    nodeLeft.setSchema(&_btreeSchema, _keyLen);
			    
			    ListT<CegoBTreeNode> modNodeList;

			    if ( nodeLeft.propagate(propValue, propLeft, propRight, modNodeList) == false )
			    {
								
				propLeft = nodeLeft;
				
				// since we still have to add the right node, save it
				CegoBTreeNode addNode = propRight;

				// create new 
				CegoBufferPage newPage; 
				_pObjMng->getNewFilePage(newPage, _tabSetId, CegoObject::BTREE, true, doAppend);
				newPage.initPage(CegoBufferPage::BTREE_NODE);
				fixedPageStack.Push(newPage);
				
				propRight.setType(CegoBTreeNode::NODE);
				propRight.setPtr(newPage.getChunkEntry(), newPage.getChunkLen());
				propRight.initNode();
				propRight.setFileId(newPage.getFileId());
				propRight.setPageId(newPage.getPageId());
				propRight.setSchema(&_btreeSchema, _keyLen);
				
				propLeft.split(propRight);
				propValue = propLeft.getMax();
				
				// cout << "!!!!!!!!!!! Decrease entries for page " << nodeLeft.getPageId() << " !!!!!!!!!" << endl;
				// propLeft.decEntries();
				// cout << "!!!!!!!!!!! New value is  " << nodeLeft.numEntries() << " !!!!!!!!!" << endl;
				
				// add new node
				CegoBTreeValue iv1 = addNode.getMin();
				CegoBTreeValue iv2 = propRight.getMin();
				
				if ( iv1.isHigher(iv2, &_btreeSchema) )
				{
				    // cout << "Adding new node on right !!!!" << endl;
				    propRight.addNode(iv1, addNode, modNodeList);
				}
				else
				{
				    // cout << "Adding new node on left !!!!" << endl;
				    propLeft.addNode(iv1, addNode, modNodeList);
				}
				
			    }
			    else
			    {
				// cout << "Node page after propagation " << endl;
				// propNode.printNode();

				/*

				if ( leafLeft.getNextFileId() && leafLeft.getNextPageId() )
				{
				    leafRight.setNextFileId(leafLeft.getNextFileId());
				    leafRight.setNextPageId(leafLeft.getNextPageId());
				    newPage.setNextFileId(leafLeft.getNextFileId());
				    newPage.setNextPageId(leafLeft.getNextPageId());
				}
				
				leafLeft.setNextFileId(newPage.getFileId());
				leafLeft.setNextPageId(newPage.getPageId());
				bp.setNextFileId(newPage.getFileId());
				bp.setNextPageId(newPage.getPageId());
				*/
				

				isPropagated = true;
			    }

			    // set up next page information

			    
			    CegoBTreeNode* pModNode = modNodeList.First();
			    while ( pModNode )
			    {

				// cout << "A Setting up post node for node " << pModNode->getFileId() << " " << pModNode->getPageId()
				//      << " to value " << pModNode->getNextFileId() << " " << pModNode->getNextPageId() << endl;

				CegoBufferPage modPage;				

				_pDBMng->bufferFix(modPage, _tabSetId, pModNode->getFileId(), pModNode->getPageId(), CegoBufferPool::NOSYNC, _pObjMng->getLockHandler());
				modPage.setNextFileId(pModNode->getNextFileId());
				modPage.setNextPageId(pModNode->getNextPageId());
				_pDBMng->bufferUnfix(modPage, true, _pObjMng->getLockHandler());	
				
				pModNode = modNodeList.Next();
			    }

			    
			}
			else // root node reached, adding new root
			{
			    
			    // cout << "Adding new root page .." << endl;
			    
			    CegoBufferPage rootPage; 
			    _pObjMng->getNewFilePage(rootPage, _tabSetId, CegoObject::BTREE, true, doAppend);
			    rootPage.initPage(CegoBufferPage::BTREE_NODE);
			   
			    if ( isFirst && _rootFixed )
			    {
				// cout << "Adding _rootBP to stack" << _rootBP.getFileId() << " "  << _rootBP.getPageId() << endl;
				fixedPageStack.Push(_rootBP);				
			    }
			    else
			    {
				// cout << "Adding rootPage to stack" << endl; 
				fixedPageStack.Push(rootPage);				
			    }

			    _rootBP = rootPage;

			    CegoBTreeNode rootNode;
			    rootNode.setType(CegoBTreeNode::NODE);
			    rootNode.setPtr(rootPage.getChunkEntry(), rootPage.getChunkLen());
			    rootNode.initNode();			
			    rootNode.setFileId(rootPage.getFileId());
			    rootNode.setPageId(rootPage.getPageId());
			    rootNode.setSchema(&_btreeSchema, _keyLen);

			    ListT<CegoBTreeNode> modNodeList;			    
			    rootNode.propagate(propValue, propLeft, propRight, modNodeList);
			 

			    CegoBTreeNode* pModNode = modNodeList.First();
			    while ( pModNode )
			    {

				// cout << "B Setting up post node for node " << pModNode->getFileId() << " " << pModNode->getPageId()
				//     << " to value " << pModNode->getNextFileId() << " " << pModNode->getNextPageId() << endl;

				CegoBufferPage modPage;				

				_pDBMng->bufferFix(modPage, _tabSetId, pModNode->getFileId(), pModNode->getPageId(), CegoBufferPool::NOSYNC, _pObjMng->getLockHandler());
				modPage.setNextFileId(pModNode->getNextFileId());
				modPage.setNextPageId(pModNode->getNextPageId());
				_pDBMng->bufferUnfix(modPage, true, _pObjMng->getLockHandler());	
				
				pModNode = modNodeList.Next();
			    }


   
			    // rootNode.printNode();
			    
			    _pBTO->setDataFileId(rootPage.getFileId());
			    _pBTO->setDataPageId(rootPage.getPageId());
			    
			    char *p;
			    int len;
			    unsigned long lockId = _pObjMng->claimDataPtr(_tabSetId, 
									  CegoLockHandler::WRITE, 
									  CegoBufferPool::NOSYNC,
									  sysEntry, p, len);
			    _pBTO->encode(p);
			    _pObjMng->releaseDataPtr(lockId, true);
			    
			    isPropagated=true;
			}
		    }		
		}
		else
		{
		    if ( bp != _rootBP || _rootFixed == false)
			_pDBMng->bufferUnfix(bp, true, _pObjMng->getLockHandler());
		}
		
		if ( duplicateAdded == false )
		{
		    _pBTO->increaseRelevance();

		    char *p;
		    int len;
		    unsigned long lockId = _pObjMng->claimDataPtr(_tabSetId, 
								  CegoLockHandler::WRITE, 
								  CegoBufferPool::NOSYNC,
								  sysEntry, p, len);
		    _pBTO->encode(p);
		    _pObjMng->releaseDataPtr(lockId, true);

		}

		isInserted = true;	    
	    }

	    isFirst = false;
	}
    }
    catch ( Exception e )
    {
	CegoBufferPage bp;
	while ( parentPageStack.Pop(bp) )
	{
	    if ( bp != _rootBP || _rootFixed == false)
	    {		
		_pDBMng->bufferUnfix(bp, true, _pObjMng->getLockHandler());
	    }
	}
	
	while ( fixedPageStack.Pop(bp) )
	{
	    if ( bp != _rootBP || _rootFixed == false)
	    {	     
		_pDBMng->bufferUnfix(bp, true, _pObjMng->getLockHandler());
	    }   
	} 

	throw e;	
    }

    // cout << "Unfixing pages .." << endl;

    // unfix node pages on stack
    CegoBufferPage bp;
    while ( parentPageStack.Pop(bp) )
    {
	if ( bp != _rootBP || _rootFixed == false )
	{
	    // cout << "Unfixing parent page .." << bp.getFileId() << " "  << bp.getPageId() << endl;
	    _pDBMng->bufferUnfix(bp, true, _pObjMng->getLockHandler());
	    // cout << "Done" << endl;
	}
    }
    
    while ( fixedPageStack.Pop(bp) )
    {
	if ( bp != _rootBP || _rootFixed == false)
	{
	    // cout << "Unfixing fixed page .." << bp.getFileId() << " "  << bp.getPageId() << endl;
	    _pDBMng->bufferUnfix(bp, true, _pObjMng->getLockHandler());
	    // cout << "Done" << endl;
	}
    }    
}

void CegoBTreeManager::dumpBTree()
{

    int fileId = _pBTO->getDataFileId();
    int pageId = _pBTO->getDataPageId();
    
    cout << "Root Page is " << fileId << " " << pageId << endl;
    
    if ( fileId == 0 && pageId == 0 )
    {
	throw Exception(EXLOC, Chain("BTree object is not valid"));
    }

    dumpNode(0, fileId, pageId);

}

void CegoBTreeManager::dumpNode(int level, int fileId, int pageId)
{

    CegoBufferPage bp;
    
    _pDBMng->bufferFix(bp, _tabSetId, fileId, pageId, CegoBufferPool::NOSYNC, _pObjMng->getLockHandler());

    CegoBTreeNode traceNode;
    traceNode.setPtr(bp.getChunkEntry(), bp.getChunkLen());	    
    traceNode.setSchema(&_btreeSchema, _keyLen);
    traceNode.setFileId(fileId);
    traceNode.setPageId(pageId);
    traceNode.setNextFileId(bp.getNextFileId());
    traceNode.setNextPageId(bp.getNextPageId());

    if ( bp.getType() == CegoBufferPage::BTREE_NODE )
    {
	traceNode.setType(CegoBTreeNode::NODE);	
	traceNode.printNode(level);
	
	int childFid, childPid;
	
	traceNode.reset();
	while ( traceNode.nextChildPointer(childFid, childPid) )
	{
	    dumpNode(level + 3, childFid, childPid);  
	}
    }
    else
    {
	traceNode.setType(CegoBTreeNode::LEAF);
	traceNode.printNode(level);
    }
   
    _pDBMng->bufferUnfix(bp, false, _pObjMng->getLockHandler());

}

void CegoBTreeManager::deleteBTree(const CegoBTreeValue &iv,
				   const CegoDataPointer& dp)
{

    
    // we have to keep the index object in buffer pool 
    CegoBufferPage bp;
    _pObjMng->getObjectWithFix(_tabSetId, _btreeName, _btreeType, *_pBTO, bp);
    
    CegoDataPointer sysEntry(bp.getFileId(), bp.getPageId(), bp.getEntryPos());

    try 
    {
	deleteBTree(sysEntry, iv, dp);
    }
    catch ( Exception e )
    {
	_pDBMng->bufferUnfix(bp, true, _pObjMng->getLockHandler());
	throw e;
    }
    _pDBMng->bufferUnfix(bp, true, _pObjMng->getLockHandler());
    
}


void CegoBTreeManager::deleteBTree(const CegoDataPointer& sysEntry, 
				   const CegoBTreeValue &iv,
				   const CegoDataPointer& dp)
{    

    int fileId = _pBTO->getDataFileId();
    int pageId = _pBTO->getDataPageId();

    if ( fileId == 0 && pageId == 0 )
    {
	throw Exception(EXLOC, Chain("BTree object is not valid"));
    }
    
    bool isDeleted = false;

    StackT<CegoBufferPage> parentPageStack;

    try
    {

	while ( ! isDeleted )
	{
	    
	    CegoBufferPage bp;
	    
	    _pDBMng->bufferFix(bp, _tabSetId, fileId, pageId, CegoBufferPool::NOSYNC, _pObjMng->getLockHandler());
	    
	    if ( bp.getType() == CegoBufferPage::BTREE_NODE )
	    {
		
		CegoBTreeNode traceNode;
		traceNode.setType(CegoBTreeNode::NODE);
		traceNode.setPtr(bp.getChunkEntry(), bp.getChunkLen());	    
		traceNode.setSchema(&_btreeSchema, _keyLen);
		traceNode.setFileId(bp.getFileId());
		traceNode.setPageId(bp.getPageId());
		
		// traceNode.printNode();
		
		traceNode.getChildPage(iv, fileId, pageId);
		
		// cout << "Tracing to child page .." << fileId << " " << pageId << endl; 
		
		parentPageStack.Push(bp);
	    }
	    else if ( bp.getType() == CegoBufferPage::BTREE_LEAF )
	    {
		
		CegoBTreeNode leaf;
		leaf.setType(CegoBTreeNode::LEAF);
		leaf.setPtr(bp.getChunkEntry(), bp.getChunkLen());
		leaf.setFileId(bp.getFileId());
		leaf.setPageId(bp.getPageId());
		leaf.setSchema(&_btreeSchema, _keyLen);
		leaf.setNextFileId(bp.getNextFileId());
		leaf.setNextPageId(bp.getNextPageId());
		
		int leafFileId = bp.getFileId();
		int leafPageId = bp.getPageId();
		int leafNextFileId = bp.getNextFileId();
		int leafNextPageId = bp.getNextPageId();
		
		while ( isDeleted == false )
		{
		    isDeleted = leaf.deleteValue(iv, dp);
		    if ( isDeleted == false )
		    {
			if ( leafNextFileId != 0 && leafNextPageId != 0 )
			{
			    _pDBMng->bufferUnfix(bp, false, _pObjMng->getLockHandler());
			    _pDBMng->bufferFix(bp, _tabSetId, leafNextFileId, leafNextPageId, CegoBufferPool::NOSYNC, _pObjMng->getLockHandler());
			    leaf.setPtr(bp.getChunkEntry(), bp.getChunkLen());
			    
			    leafFileId = bp.getFileId();
			    leafPageId = bp.getPageId();
			    leafNextFileId = bp.getNextFileId();
			    leafNextPageId = bp.getNextPageId();
			    
			    // leaf.setFileId(bp.getFileId());
			    // leaf.setPageId(bp.getPageId());			
			}
			else
			{
			    _pDBMng->bufferUnfix(bp, true, _pObjMng->getLockHandler());
			    throw Exception(EXLOC, Chain("Entry not found in btree"));
			}
		    }
		}
		
		_pDBMng->bufferUnfix(bp, true, _pObjMng->getLockHandler());
		
		// check for duplicates
		
		bool hasDuplicate=false;
		
		CegoBufferPage cbp1;
		_pDBMng->bufferFix(cbp1, _tabSetId, leafFileId, leafPageId, CegoBufferPool::NOSYNC, _pObjMng->getLockHandler());
		leaf.setPtr(cbp1.getChunkEntry(), cbp1.getChunkLen());
		if ( leaf.valueExists(iv) )
		{
		    hasDuplicate=true;
		}
		else if ( leafNextFileId != 0 && leafNextPageId != 0 )
		{
		    // now we still have to check the next following page for duplicates
		    CegoBufferPage cbp2;
		    _pDBMng->bufferFix(cbp2, _tabSetId, leafNextFileId, leafNextPageId, CegoBufferPool::NOSYNC, _pObjMng->getLockHandler());
		    leaf.setPtr(bp.getChunkEntry(), bp.getChunkLen());
		    if ( leaf.valueExists(iv) )
		    {
			hasDuplicate=true;
		    }
		    
		    _pDBMng->bufferUnfix(cbp2, false, _pObjMng->getLockHandler());
		}
		
		_pDBMng->bufferUnfix(cbp1, false, _pObjMng->getLockHandler());
		
		if ( hasDuplicate == false )
		{
		    
		    _pBTO->decreaseRelevance();
		    char *p;
		    int len;
		    unsigned long lockId = _pObjMng->claimDataPtr(_tabSetId, 
								  CegoLockHandler::WRITE, 
								  CegoBufferPool::NOSYNC,
								  sysEntry, p, len);
		    _pBTO->encode(p);
		    _pObjMng->releaseDataPtr(lockId, true);
		    
		}
	    }
	}
    }
    catch ( Exception e ) 
    {
	CegoBufferPage bp;
	while ( parentPageStack.Pop(bp) )
	{
	_pDBMng->bufferUnfix(bp, true, _pObjMng->getLockHandler());
	}
	throw e;
    }

    // unfix node pages on stack
    CegoBufferPage bp;
    while ( parentPageStack.Pop(bp) )
    {
	_pDBMng->bufferUnfix(bp, true, _pObjMng->getLockHandler());
    }
}

void CegoBTreeManager::freeBTree()
{
    freeNodePages(_pBTO->getDataFileId(), _pBTO->getDataPageId());
}


void CegoBTreeManager::freeNodePages(int fileId, int pageId)
{
    int pageCount = 0;

    CegoBufferPage bp;    
    _pDBMng->bufferFix(bp, _tabSetId, fileId, pageId, CegoBufferPool::NOSYNC, _pObjMng->getLockHandler());

    if ( bp.getType() == CegoBufferPage::BTREE_NODE )
    {

	CegoBTreeNode traceNode;
	traceNode.setPtr(bp.getChunkEntry(), bp.getChunkLen());	    
	traceNode.setSchema(&_btreeSchema, _keyLen);
	traceNode.setFileId(fileId);
	traceNode.setPageId(pageId);	
	traceNode.setType(CegoBTreeNode::NODE);	

	int childFid, childPid;
	
	traceNode.reset();
	while ( traceNode.nextChildPointer(childFid, childPid) )
	{	    
	    freeNodePages(childFid, childPid);
	}
    }

    _pDBMng->bufferRelease(bp, _pObjMng->getLockHandler());

}

int CegoBTreeManager::getNumPages()
{
    return countNodePages(_pBTO->getDataFileId(), _pBTO->getDataPageId());
}

int CegoBTreeManager::countNodePages(int fileId, int pageId)
{

    int pageCount = 0;

    CegoBufferPage bp;    
    _pDBMng->bufferFix(bp, _tabSetId, fileId, pageId, CegoBufferPool::NOSYNC, _pObjMng->getLockHandler());

    if ( bp.getType() == CegoBufferPage::BTREE_NODE )
    {

	CegoBTreeNode traceNode;
	traceNode.setPtr(bp.getChunkEntry(), bp.getChunkLen());	    
	traceNode.setSchema(&_btreeSchema, _keyLen);
	traceNode.setFileId(fileId);
	traceNode.setPageId(pageId);	
	traceNode.setType(CegoBTreeNode::NODE);	

	int childFid, childPid;
	
	traceNode.reset();
	while ( traceNode.nextChildPointer(childFid, childPid) )
	{
	    pageCount += countNodePages(childFid, childPid);
	}
    }
    else
    {
	pageCount++;
    }
   
    _pDBMng->bufferUnfix(bp, true, _pObjMng->getLockHandler());

    return pageCount;
}


